class P4Action:
    def __init__(self, name: str, uid: int) -> None:
        self.name = name
        self.uid = uid
        # List containing all the writes for a specific action
        self.register_writes = []
        self.register_reads = []
        self.prov_width = (
            0  # Total width of the prov record - size(index) + size(value)
        )

    def __str__(self) -> str:
        return f"P4Action Name: {self.name}"