#!/bin/bash
H1=$1
H2=$2
nohup ssh -t -i /home/dhiraj/CREASE-PROVP4/fabrictestbed-extensions/env/fabric_config/slice_key -F /home/dhiraj/CREASE-PROVP4/fabrictestbed-extensions/env/fabric_config/ssh_config ubuntu@${H1} "ping -c 3 ${H2}" 2>&1 &