"use strict";(self.webpackChunkjupyter_cytoscape=self.webpackChunkjupyter_cytoscape||[]).push([[480],{4480:(e,t,s)=>{s.r(t),s.d(t,{default:()=>i});var a=s(4217),n=s(8657);const i={id:"jupyter-cytoscape:plugin",requires:[a.IJupyterWidgetRegistry],activate:function(e,t){t.registerWidget({name:n.o,version:n.Y,exports:()=>s.e(106).then(s.bind(s,3106))})},autoStart:!0}},8657:(e,t,s)=>{s.d(t,{Y:()=>n,o:()=>i});const a=s(4147),n=a.version,i=a.name},4147:e=>{e.exports=JSON.parse('{"name":"jupyter-cytoscape","version":"1.3.3","description":"A Cytoscape Jupyter Widget","keywords":["jupyter","jupyterlab","jupyterlab-extension","widgets","cytoscape"],"files":["{dist,lib}/**/*.{js,ts,map}","css/*.css","LICENSE"],"homepage":"https://github.com/cytoscape/ipycytoscape","bugs":{"url":"https://github.com/cytoscape/ipycytoscape/issues"},"license":"BSD-3-Clause","author":{"name":"Mariana Meireles","email":"mariana.meireles@quantstack.net"},"main":"lib/index.js","types":"./lib/index.d.ts","repository":{"type":"git","url":"https://github.com/cytoscape/ipycytoscape"},"scripts":{"build:prod":"npm run build:lib && npm run build:all","build":"npm run build:lib && npm run build:all:dev","build:labextension":"npm run clean:labextension && jupyter labextension build .","build:labextension:dev":"npm run clean:labextension && jupyter labextension build --development True .","build:lib":"tsc","build:nbextension":"webpack --mode=production","build:nbextension:dev":"webpack --mode=development","build:all":"npm run build:labextension && npm run build:nbextension","build:all:dev":"npm run build:labextension:dev && npm run build:nbextension:dev","clean":"npm run clean:lib && npm run clean:nbextension","clean:lib":"rimraf lib","clean:labextension":"rimraf ipycytoscape/labextension","clean:nbextension":"rimraf ipycytoscape/nbextension/static/index.js","lint":"eslint . --ext .ts --fix","lint-check":"eslint . --ext .ts","prepack":"npm run build:lib","test":"npm run test:firefox","test:chrome":"karma start --browsers=Chrome tests/karma.conf.js","test:debug":"karma start --browsers=Chrome --singleRun=false --debug=true tests/karma.conf.js","test:firefox":"karma start --browsers=Firefox tests/karma.conf.js","test:ie":"karma start --browsers=IE tests/karma.conf.js","watch":"npm-run-all -p watch:*","watch:lib":"tsc -w","watch:nbextension":"webpack --watch --mode=development","watch:labextension":"jupyter labextension watch ."},"dependencies":{"@jupyter-widgets/base":"^1.1.10 || ^2 || ^3 || ^4.0.0","@types/cytoscape":"^3.8.8","cytoscape":"^3.14.0","cytoscape-cola":"^2.3.0","cytoscape-dagre":"^2.2.2","cytoscape-klay":"^3.1.3","cytoscape-popper":"^1.0.6","tippy.js":"^5.2.1"},"devDependencies":{"@jupyterlab/builder":"^3.0.0","@phosphor/application":"^1.6.0","@phosphor/widgets":"^1.6.0","@types/expect.js":"^0.3.29","@types/mocha":"^5.2.5","@types/node":"^10.11.6","@types/webpack-env":"^1.13.6","@typescript-eslint/eslint-plugin":"^5.10.0","@typescript-eslint/parser":"^5.10.0","css-loader":"^3.2.0","eslint":"^8.0.0","eslint-config-prettier":"^6.11.0","eslint-plugin-prettier":"^3.1.3","expect.js":"^0.3.1","fs-extra":"^7.0.0","karma":"^6.0.0","karma-chrome-launcher":"^2.2.0","karma-firefox-launcher":"^1.1.0","karma-ie-launcher":"^1.0.0","karma-mocha":"^2.0.1","karma-mocha-reporter":"^2.2.5","karma-typescript":"^5.5.0","lint-staged":"^10.2.2","mkdirp":"^0.5.1","mocha":"^9.2.2","npm-run-all":"^4.1.3","prettier":"^2.0.5","rimraf":"^2.6.2","source-map-loader":"^0.2.4","style-loader":"^1.0.0","ts-loader":"^5.2.1","typescript":"~3.7.1","webpack":"^5.0.0","webpack-cli":"^4.0.0"},"jupyterlab":{"extension":"lib/plugin","outputDir":"ipycytoscape/labextension","sharedPackages":{"@jupyter-widgets/base":{"bundled":false,"singleton":true}}},"prettier":{"singleQuote":true}}')}}]);