"use strict";(self.webpackChunkjupyter_leaflet=self.webpackChunkjupyter_leaflet||[]).push([[176],{6176:(t,e,s)=>{s.r(e),s.d(e,{Spinner:()=>o});var n=function(){return n=Object.assign||function(t){for(var e,s=1,n=arguments.length;s<n;s++)for(var i in e=arguments[s])Object.prototype.hasOwnProperty.call(e,i)&&(t[i]=e[i]);return t},n.apply(this,arguments)},i={lines:12,length:7,width:5,radius:10,scale:1,corners:1,color:"#000",fadeColor:"transparent",animation:"spinner-line-fade-default",rotate:0,direction:1,speed:1,zIndex:2e9,className:"spinner",top:"50%",left:"50%",shadow:"0 0 1px transparent",position:"absolute"},o=function(){function t(t){void 0===t&&(t={}),this.opts=n(n({},i),t)}return t.prototype.spin=function(t){return this.stop(),this.el=document.createElement("div"),this.el.className=this.opts.className,this.el.setAttribute("role","progressbar"),this.el.style.position=this.opts.position,this.el.style.width="0",this.el.style.zIndex=this.opts.zIndex.toString(),this.el.style.left=this.opts.left,this.el.style.top=this.opts.top,this.el.style.transform="scale(".concat(this.opts.scale,")"),t&&t.insertBefore(this.el,t.firstChild||null),function(t,e){var s=Math.round(e.corners*e.width*500)/1e3+"px",n="none";!0===e.shadow?n="0 2px 4px #000":"string"==typeof e.shadow&&(n=e.shadow);for(var i=function(t){for(var e=/^\s*([a-zA-Z]+\s+)?(-?\d+(\.\d+)?)([a-zA-Z]*)\s+(-?\d+(\.\d+)?)([a-zA-Z]*)(.*)$/,s=[],n=0,i=t.split(",");n<i.length;n++){var o=i[n].match(e);if(null!==o){var r=+o[2],a=+o[5],l=o[4],h=o[7];0!==r||l||(l=h),0!==a||h||(h=l),l===h&&s.push({prefix:o[1]||"",x:r,y:a,xUnits:l,yUnits:h,end:o[8]})}}return s}(n),o=0;o<e.lines;o++){var l=~~(360/e.lines*o+e.rotate),h=document.createElement("div");h.style.position="absolute",h.style.top="".concat(-e.width/2,"px"),h.style.width=e.length+e.width+"px",h.style.height=e.width+"px",h.style.background=r(e.fadeColor,o),h.style.borderRadius=s,h.style.transformOrigin="left",h.style.transform="rotate(".concat(l,"deg) translateX(").concat(e.radius,"px)");var p=o*e.direction/e.lines/e.speed;p-=1/e.speed;var d=document.createElement("div");d.style.width="100%",d.style.height="100%",d.style.background=r(e.color,o),d.style.borderRadius=s,d.style.boxShadow=a(i,l),d.style.animation="".concat(1/e.speed,"s linear ").concat(p,"s infinite ").concat(e.animation),h.appendChild(d),t.appendChild(h)}}(this.el,this.opts),this},t.prototype.stop=function(){return this.el&&(this.el.parentNode&&this.el.parentNode.removeChild(this.el),this.el=void 0),this},t}();function r(t,e){return"string"==typeof t?t:t[e%t.length]}function a(t,e){for(var s=[],n=0,i=t;n<i.length;n++){var o=i[n],r=l(o.x,o.y,e);s.push(o.prefix+r[0]+o.xUnits+" "+r[1]+o.yUnits+o.end)}return s.join(", ")}function l(t,e,s){var n=s*Math.PI/180,i=Math.sin(n),o=Math.cos(n);return[Math.round(1e3*(t*o+e*i))/1e3,Math.round(1e3*(-t*i+e*o))/1e3]}}}]);